from typing import List, Dict

import pandas as pd

from nytimes_scraper.comments.util import flatten_replies, remove_reply_references


def comments_to_df(comments: List[Dict]) -> pd.DataFrame:
    all_comments = remove_reply_references(flatten_replies(comments))
    df = pd.json_normalize(all_comments).set_index('commentID')

    if 'parentID' in df:
        df['parentID'] = df['parentID'].astype(pd.Int64Dtype())

    for col in ['status', 'commentType']:
        if col in df:
            df[col] = df[col].astype(pd.CategoricalDtype())

    for col in ['createDate', 'updateDate', 'approveDate']:
        if col in df:
            df[col] = pd.to_datetime(df[col], unit='s')

    return df



from typing import List, Dict


def flatten_replies(comments: List[Dict]) -> List[Dict]:
    """Flattens all comments and replies into one list without copying or modifying the objects"""

    result = []
    for comment in comments:
        result.append(comment)
        result.extend(flatten_replies(comment['replies']))

    return result


def remove_reply_references(comments: List[Dict]) -> List[Dict]:
    """Removes the `replies` list from every comments"""

    result = []
    for comment in comments:
        copy = comment.copy()
        copy.pop('replies', None)
        result.append(copy)

    return result



from typing import List, Dict

import pandas as pd


def articles_to_df(articles: List[Dict]) -> pd.DataFrame:
    df = pd.json_normalize(articles).set_index('_id')

    if 'pub_date' in df:
        df['pub_date'] = pd.to_datetime(df['pub_date'])

    for col in ['document_type', 'news_desk', 'section_name', 'type_of_material', 'byline.organization', 'subsectoinName']:
        if col in df:
            df[col] = df[col].astype(pd.CategoricalDtype())

    return df[~df.index.duplicated()]



import datetime as dt
from pathlib import Path
from typing import Tuple, Callable

import pandas as pd

from nytimes_scraper.articles import fetch_articles_by_month, articles_to_df
from nytimes_scraper.comments import fetch_comments, comments_to_df
from nytimes_scraper.nyt_api import NytApi


out_dir = Path.cwd()


def run_scraper(api_key: str):
    """Scrape articles and comments month by month, starting from the current month"""

    # the year and month to be fetched
    date = dt.datetime.now().date().replace(day=1)
    while True:
        scrape_month(api_key, date)

        # go one month back
        date = (date - dt.timedelta(days=1)).replace(day=1)


def scrape_month(api_key: str, date: dt.date, force_fetch: bool = False, store: bool = True) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Scrape articles and comments for a given month. The `date.day` is ignored."""

    api = NytApi(api_key)

    article_df = cached(
        fetch=lambda: articles_to_df(fetch_articles_by_month(api, date)),
        file=out_file(date, 'articles'),
        force_fetch=force_fetch,
        store=store,
    )

    article_ids_and_urls = list(article_df['web_url'].iteritems())
    comment_df = cached(
        fetch=lambda: comments_to_df(fetch_comments(api, article_ids_and_urls)),
        file=out_file(date, 'comments'),
        force_fetch=force_fetch,
        store=store,
    )

    return article_df, comment_df


def cached(file: Path, fetch: Callable[[], pd.DataFrame], force_fetch: bool, store: bool) -> pd.DataFrame:
    if not file.exists() or force_fetch:
        df = fetch()

        if store:
            file.parent.mkdir(exist_ok=True)
            df.to_pickle(str(file))

        return df
    else:
        return pd.read_pickle(str(file))


def out_file(date: dt.date, kind: str) -> Path:
    return out_dir / f'{date.year}-{date.month:02d}-{kind}.pickle'

